import { Timing } from 'necessary-utils-core';

class QueueSettings {
  _queueRestartTime = 100;

  get QueueRestartTime() {
    return this._queueRestartTime;
  }

  set QueueRestartTime(value) {
    if (value < 100) this._queueRestartTime = 100;else this._queueRestartTime = value;
  }

  QueueCoolDownTime = 0;

  constructor() {}

}

/**
 * # Simple Queue System
 * ### Methods
 * - PushMessage
 * - Comsume
 * - ACK
 * - Clear
 * @example
 * `Define your queue`
 * new SimpleQueue<MessageObject>();
 * @example
 * `Publish a message`
 * new SimpleQueue<MessageObject>().Push(MessageObject);
 * @example
 * `Full example`
 * var queue = new SimpleQueue<MessageObject>();
 * queue.Consume(async (message: MessageObject) => {
 *  do something...
 * });
 * queue.Push(MessageObject);
 */

class BasicQueue {
  _messages = [];
  Handler = null;
  AutoACK = false;
  _queueStarted = false;
  _currentMessage = null;
  Settings;

  constructor(settings = null) {
    if (settings === null) this.Settings = new QueueSettings();else this.Settings = settings;
  }

  Ack() {
    if (this._currentMessage === null || this.Handler == null) return;

    const i = this._messages.indexOf(this._currentMessage);

    if (i !== -1) this._messages.splice(i, 1);
    this._currentMessage = null;
    this._queueStarted = this._messages.length > 0;
  }

  PushMessage(message) {
    this._messages.push(message);

    this.start();
  }

  Consume(handler, autoAck) {
    this.Handler = handler;
    this.AutoACK = autoAck;
  }

  get MessageCount() {
    return this._messages.length;
  }

  Clear() {
    let removedMessages = this._messages.filter(x => !x.UnAck);

    this._messages = this._messages.filter(x => !removedMessages.includes(x));
    return removedMessages;
  }

  Stop() {
    this._queueStarted = false;
  }

  Intercept = {
    WhenZeroMessage: () => {
      if (this.MessageCount === 0) throw new Error("Zero message count. Process was intercepted.");
      return this.Intercept;
    },
    WhenAnyMessage: () => {
      if (this.MessageCount > 0) throw new Error("Queue contains message. Process was intercepted.");
      return this.Intercept;
    },
    When: async handler => {
      let currentMessage = Object.assign({}, this._currentMessage);
      const result = await handler(this.MessageCount, currentMessage);
      if (result !== true) throw new Error("Custom handler returns false. Process was intercepted.");
      return this.Intercept;
    },
    WhenQueueStarted: () => {
      if (this._queueStarted) throw new Error("Queue processing right now. Process was intercepted.");
      return this.Intercept;
    },
    WhenQueueStopped: () => {
      if (!this._queueStarted) throw new Error("Queue stopped. Process was intercepted.");
      return this.Intercept;
    }
  };

  async start() {
    if (this._queueStarted) return;
    this._queueStarted = true;

    while (this._queueStarted) {
      this._currentMessage = this._messages[0];

      if (this.Handler === null || this._currentMessage.UnAck === true) {
        await Timing.Sleep(this.Settings.QueueRestartTime);
        continue;
      }

      await Timing.Sleep(this.Settings.QueueCoolDownTime);

      try {
        this._currentMessage.UnAck = true;
        await this.Handler(this._currentMessage);
      } catch (error) {
        console.error("Consumer Error Ocurred. See the details; \n", error);
      }

      if (this.AutoACK) this.Ack();
    }
  }

}

/**
 * # DEPRECATED
 *
 * This working like Task.Delay() on the C#
 * @param ms
 * @returns
 * @example `
 *  doSomething...
 *  wait(1000) // One second
 *  continue...
 *  `
 */
const wait = ms => {
  return new Promise(resolve => setTimeout(resolve, ms));
};
const uuidv4 = () => {
  return "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g, function (c) {
    var r = Math.random() * 16 | 0,
        v = c == "x" ? r : r & 0x3 | 0x8;
    return v.toString(16);
  });
};

class BasicQueueMessage {
  Id;
  Date;
  UnAck = false;

  constructor() {
    this.Id = uuidv4();
    this.Date = new Date().getTime();
  }

}

export { BasicQueue, BasicQueueMessage, QueueSettings, uuidv4, wait };
